# -*- encoding: utf-8 -*-
"""Unifont Utils - Diff"""

from typing import List, Union

from rich.console import Console
from rich.text import Text
from rich.table import Table

from .base import Validator as V
from .converter import Converter as C
from .glyphs import Glyph


def get_img_data(glyph: Union[str, Glyph]) -> List[int]:
    """Input a `.hex` string or a Glyph object and output its image data.

    Args:
        glyph (Union[str, Glyph]): The glyph to be converted.

    Returns:
        List[int]: The converted image data.

    Raises:
        TypeError: If the glyph is not a valid type.
    """

    if isinstance(glyph, str):
        glyph = V.hex_str(glyph)
        width = 16 if len(glyph) == 64 else 8
        return C.to_img_data(glyph, width)
    if isinstance(glyph, Glyph):
        return glyph.data
    raise TypeError("The glyph must be either a .hex string or a Glyph object.")


def diff_glyphs(glyph_a: Union[str, Glyph], glyph_b: Union[str, Glyph]) -> List[int]:
    """Compares two glyphs and returns a list of differences.

    Args:
        glyph_a (Union[str, Glyph]): The first glyph to compare.
        glyph_b (Union[str, Glyph]): The second glyph to compare.

    Returns:
        List[int]: A list of differences between the two glyphs.

    Raises:
        ValueError: If the two glyphs have different sizes.
    """

    a, b = get_img_data(glyph_a), get_img_data(glyph_b)
    if len(a) != len(b):
        raise ValueError("The two glyphs must have the same size.")

    diff_list = []
    for i, j in zip(a, b):
        if i and not j:
            diff_list.append("-")
        elif not i and j:
            diff_list.append("+")
        elif i and j:
            diff_list.append("1")
        else:
            diff_list.append("0")

    return diff_list


def print_diff(
    glyph_a: Union[str, Glyph],
    glyph_b: Union[str, Glyph],
    *,
    black_and_white: bool = True,
) -> None:
    """Prints the differences between two glyphs.

    Args:
        glyph_a (Union[str, Glyph]): The first glyph to compare.
        glyph_b (Union[str, Glyph]): The second glyph to compare.
        black_and_white (bool, optional): Whether it is a black and white image.

            Defaults to `True`.

            If `True`, `0` is white and `1` is black.

            If `False`, `0` is transparent and `1` is white.
    """

    diff_list = diff_glyphs(glyph_a, glyph_b)
    a, b = get_img_data(glyph_a), get_img_data(glyph_b)
    console = Console()

    white_block = "white on white"
    black_block = "black on black"

    if black_and_white:
        white_block, black_block = black_block, white_block

    width = len(a) // 16

    def get_row(i: int, data: List[int]) -> Text:
        row_text = Text()
        row_data = data[i * width : (i + 1) * width]
        for pixel in row_data:
            block_style = white_block if pixel else black_block
            row_text.append("  ", style=block_style)
        return row_text

    def get_row_diff(i: int) -> Text:
        row_text = Text()
        row_diff = {
            "+": "green on green",
            "-": "red on red",
            "1": white_block,
            "0": black_block,
        }
        for element in diff_list[i * width : (i + 1) * width]:
            block_style = row_diff.get(str(element), black_block)
            row_text.append("  ", style=block_style)
        return row_text

    table = Table(show_lines=False, expand=False)
    table.add_column("Input")
    table.add_column("Diff")
    table.add_column("Output")

    for i in range(16):
        table.add_row(get_row(i, a), get_row_diff(i), get_row(i, b))

    console.print(table)



# -*- encoding: utf-8 -*-
"""Unifont Utils - Editor"""

from typing import Tuple

from rich.text import Text
from rich.panel import Panel
from rich.console import Group
from textual import events
from textual.app import App, ComposeResult
from textual.widgets import Header, Footer, Static
from textual.reactive import reactive

from .glyphs import Glyph, SearchPattern, ReplacePattern


class EditWidget(Static, can_focus=True):
    """Widget to display and edit a Glyph."""

    cursor_x = reactive(0)
    cursor_y = reactive(0)
    glyph: Glyph

    BINDINGS = [
        ("w,up", "move_up", "Up"),
        ("s,down", "move_down", "Down"),
        ("a,left", "move_left", "Left"),
        ("d,right", "move_right", "Right"),
        ("space", "toggle_glyph", "Toggle 0/1"),
        ("q", "quit", "Quit"),
    ]

    def __init__(self, glyph: Glyph) -> None:
        super().__init__()
        self.glyph = glyph

    def on_mount(self) -> None:
        """Actions that are executed when the widget is mounted."""
        self.render_glyph()
        self.focus()

    def watch_cursor_x(self) -> None:
        """Watch for changes to the `cursor_x` or `cursor_y` attribute."""
        self.render_glyph()

    watch_cursor_y = watch_cursor_x

    def render_glyph(self) -> None:
        """Render the glyph with the current cursor position."""

        def get_color(i: int) -> str:
            """Get color based on index."""
            return "auto" if i % 2 == 0 else ("green" if i > 9 else "red")

        def get_pixel_color(value: int) -> str:
            if value:
                return "white" if self.app.dark else "black"
            return "black" if self.app.dark else "white"

        def get_block_style(is_cursor: bool, data_value: int) -> str:
            """Get style for the glyph block."""
            pixel = get_pixel_color(data_value)
            return f"{'red' if is_cursor else pixel} on {pixel}"

        def get_nums(i: int) -> str:
            """Get the hexadecimal representation of index."""
            return hex(i)[2:].rjust(2).upper()

        width = self.glyph.width
        glyph = Text("\n  ")
        # Columns
        for i in range(width):
            glyph.append(get_nums(i), style=f"{get_color(i)} bold")
        glyph.append("\n")

        for i in range(16):
            # Rows
            glyph.append(f"{get_nums(i)} ", style=f"{get_color(i)} bold")
            # Glyph pixel blocks
            for j in range(width):
                is_cursor = self.cursor_x == j % width and self.cursor_y == i
                block_style = get_block_style(is_cursor, self.glyph.data[i * width + j])
                char = "⬥ " if is_cursor else "  "
                glyph.append(char, style=block_style)
            glyph.append("\n")

        position = Text(
            f"Position: ({self.cursor_x}, {self.cursor_y})",
            justify="center",
            style="bold",
        )
        panel = Panel(
            glyph,
            title=f"U+{self.glyph.code_point} ({self.glyph.character})",
            subtitle=position,
        )

        self.update(
            Group(Text(self.glyph.unicode_name, justify="center", style="bold"), panel)
        )

    def _get_index(self) -> int:
        """Get the index of the current cursor position."""
        return self.cursor_y * self.glyph.width + self.cursor_x

    def _move_cursor(self, dx: int, dy: int) -> None:
        """Move cursor by the given offsets."""
        self.cursor_x = min(max(0, self.cursor_x + dx), self.glyph.width - 1)
        self.cursor_y = min(max(0, self.cursor_y + dy), 15)

    def action_move_up(self) -> None:
        """Move the cursor up."""
        self._move_cursor(0, -1)

    def action_move_down(self) -> None:
        """Move the cursor down."""
        self._move_cursor(0, 1)

    def action_move_left(self) -> None:
        """Move the cursor left."""
        self._move_cursor(-1, 0)

    def action_move_right(self) -> None:
        """Move the cursor right."""
        self._move_cursor(1, 0)

    def action_toggle_glyph(self) -> None:
        """Toggle the glyph's visibility."""
        index = self._get_index()
        self.glyph.update_data_at_index(index, int(not self.glyph.data[index]))
        self.render_glyph()

    def action_quit(self) -> None:
        """Quit the application."""
        self.app.exit()

    def _handle_mouse_event(self, event, update_data: bool = False) -> None:
        """Handle mouse events for click and movement."""

        grid_x = (event.x - 5) // 2
        grid_y = event.y - 4

        if 0 <= grid_x < self.glyph.width and 0 <= grid_y < 16:
            self.cursor_x = grid_x
            self.cursor_y = grid_y

            index = self._get_index()
            if update_data:
                if event.button == 1 and not event.ctrl:
                    self.glyph.update_data_at_index(index, 1)
                elif event.button == 3 or (event.button == 1 and event.ctrl):
                    self.glyph.update_data_at_index(index, 0)

            self.render_glyph()

    def on_mouse_down(self, event: events.MouseDown) -> None:
        """Handle mouse click events."""
        self._handle_mouse_event(event, update_data=True)

    def on_mouse_move(self, event: events.MouseMove) -> None:
        """Handle mouse move events."""
        self._handle_mouse_event(event, update_data=bool(event.button))


class ReplaceWidget(Static, can_focus=True):
    """Widget to display and edit a Glyph."""

    match_index = reactive(0)
    matches = []
    glyph: Glyph

    BINDINGS = [
        ("a,left", "prev", "Previous"),
        ("d,right", "next", "Next"),
        ("space", "apply", "Apply"),
        ("q", "quit", "Quit"),
    ]

    def __init__(
        self,
        glyph: Glyph,
        search_pattern: SearchPattern,
        replace_pattern: ReplacePattern,
    ) -> None:
        super().__init__()
        self.glyph = glyph
        self.search_pattern = search_pattern
        self.replace_pattern = replace_pattern
        self.matches = self.glyph.find_matches(search_pattern)

    def on_mount(self) -> None:
        """Actions that are executed when the widget is mounted."""
        self.render_glyph()
        self.focus()

    def watch_match_index(self) -> None:
        """Watch for changes to the `match_index` attribute."""
        self.render_glyph()

    def render_glyph(self) -> None:
        """Render the glyph with the current cursor position."""

        def get_color(i: int) -> str:
            """Get color based on index."""
            return "auto" if i % 2 == 0 else ("green" if i > 9 else "red")

        def get_pixel_color(value: int) -> str:
            if value == 1:
                return "white" if self.app.dark else "black"
            return "black" if self.app.dark else "white"

        def get_block_style(i: int, j: int, current_match: Tuple[int, int]) -> str:

            width = self.glyph.width
            h = self.replace_pattern.height
            w = self.replace_pattern.width
            x, y = current_match

            if current_match and x <= i < x + h and y <= j < y + w:
                pixel = self.replace_pattern.data[(i - x) * w + (j - y)]
                pixel_color = get_pixel_color(pixel)
                if pixel == 1:
                    return "green on green"
                return f"{pixel_color} on {pixel_color}"
            pixel_color = get_pixel_color(self.glyph.data[i * width + j])
            return f"{pixel_color} on {pixel_color}"

        def get_nums(i: int) -> str:
            """Get the hexadecimal representation of index."""
            return hex(i)[2:].rjust(2).upper()

        width = self.glyph.width
        glyph = Text("\n  ")
        # Columns
        for i in range(width):
            glyph.append(get_nums(i), style=f"{get_color(i)} bold")
        glyph.append("\n")

        current_match = self.matches[self.match_index]

        for i in range(16):
            # Rows
            glyph.append(f"{get_nums(i)} ", style=f"{get_color(i)} bold")
            for j in range(width):
                block_style = get_block_style(i, j, current_match)
                glyph.append("  ", style=block_style)
            glyph.append("\n")

        match_index_text = Text(
            f"Matches ({self.match_index + 1} / {len(self.matches)})",
            justify="center",
            style="bold",
        )
        panel = Panel(
            glyph,
            title=f"U+{self.glyph.code_point} ({self.glyph.character})",
            subtitle=match_index_text,
        )

        self.update(
            Group(Text(self.glyph.unicode_name, justify="center", style="bold"), panel)
        )

    def action_prev(self) -> None:
        """Move the cursor to the previous match."""
        if self.matches:
            self.match_index = (self.match_index - 1) % len(self.matches)

    def action_next(self) -> None:
        """Move the cursor to the next match."""
        if self.matches:
            self.match_index = (self.match_index + 1) % len(self.matches)

    def action_apply(self) -> None:
        """Apply the replacement to the glyph at the current match."""
        if self.matches:
            match_pos = self.matches[self.match_index]
            self.glyph.apply_pattern(match_pos[0], match_pos[1], self.replace_pattern)
            self.render_glyph()
        self.matches = self.glyph.find_matches(self.search_pattern)

    def action_quit(self) -> None:
        """Quit the application."""
        self.app.exit()


class GlyphEditor(App):
    """Main application for editing a glyph."""

    glyph: Glyph
    CSS_PATH = "editor.tcss"

    BINDINGS = [("ctrl+d", "toggle_dark", "Toggle Dark Mode")]

    def __init__(self, glyph: Glyph) -> None:
        super().__init__()
        self.glyph = glyph
        self.edit_widget = None

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.dark = not self.dark
        if self.edit_widget:
            self.edit_widget.render_glyph()

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header()
        yield Footer()
        self.edit_widget = EditWidget(self.glyph)
        yield self.edit_widget


class GlyphReplacer(App):
    """Main application for replacing patterns in a glyph."""

    glyph: Glyph
    CSS_PATH = "editor.tcss"

    BINDINGS = [("ctrl+d", "toggle_dark", "Toggle Dark Mode")]

    def __init__(
        self,
        glyph: Glyph,
        search_pattern: SearchPattern,
        replace_pattern: ReplacePattern,
    ) -> None:
        super().__init__()
        self.glyph = glyph
        self.search_pattern = search_pattern
        self.replace_pattern = replace_pattern
        self.replace_widget = None

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.dark = not self.dark
        if self.replace_widget:
            self.replace_widget.render_glyph()

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header()
        yield Footer()
        self.replace_widget = ReplaceWidget(
            self.glyph, self.search_pattern, self.replace_pattern
        )
        yield self.replace_widget



# -*- encoding: utf-8 -*-
"""Unifont Utils - Base Module"""

from pathlib import Path
from typing import List, Set, Optional, Union, TypeAlias
from collections.abc import Sequence

# Type aliases
FilePath: TypeAlias = Union[str, Path]
CodePoint: TypeAlias = Union[str, int]
CodePoints: TypeAlias = Union[Sequence[CodePoint], Set[CodePoint]]


class Validator:
    """Class for validators."""

    @staticmethod
    def code_point(code_point: CodePoint) -> str:
        """Validate a code point string and return its normalized form.

        Args:
            code_point (CodePoint): The code point string to validate.

        Returns:
            str: The normalized code point if valid.

        Raises:
            TypeError: If the code point is not a string or an integer.
            ValueError: If the code point is invalid.
        """

        if isinstance(code_point, int):
            code_point = hex(code_point)[2:]
        elif isinstance(code_point, str):
            pass
        else:
            raise TypeError("Invalid code point type. Must be a string or integer.")

        if not code_point.isalnum() or int(code_point, 16) > 0x10FFFF:
            raise ValueError(f"Invalid code point: {code_point}.")

        code_point = code_point.upper()

        for c in code_point:
            if c not in "0123456789ABCDEF":
                raise ValueError(f"Invalid character in code point: {c}.")

        return code_point.zfill(6 if len(code_point) > 4 else 4)

    @staticmethod
    def code_points(code_points: CodePoints) -> List[str]:
        """Validate acode point list and return its normalized form.

        Args:
            code_points (CodePoints): The code point string or tuple to validate.

        Returns:
            List[int]: The normalized code point list if valid.

        Raises:
            TypeError: If the code points are not a list, a set, or a `range` object.
            ValueError: If the code points are invalid.
        """

        if not isinstance(code_points, (Sequence, set)):
            raise TypeError(
                "Invalid type for the specified code points. "
                "The argument must be either a list or a range object."
            )
        if not all(isinstance(c, (str, int)) for c in code_points):
            raise TypeError("The code points in the list must be strings or integers.")

        code_points_list = [hex(int(i))[2:].zfill(4) for i in list(set(code_points))]

        return [Validator.code_point(code_point) for code_point in code_points_list]

    @staticmethod
    def hex_str(hex_str: Optional[str]) -> str:
        """Validate a hexadecimal string and return its normalized form.

        Args:
            hex_str (str, optional): The hexadecimal string to validate.

        Returns:
            str: The normalized hexadecimal string if valid.

        Raises:
            ValueError: If the hexadecimal string is invalid.
        """

        if not hex_str:
            return ""

        if len(hex_str) not in {32, 64}:
            raise ValueError(
                f"Invalid .hex string length: {hex_str} (length: {len(hex_str)})."
            )

        for c in hex_str:
            if c not in "0123456789ABCDEF":
                raise ValueError(f"Invalid character in .hex string: {c}.")

        return hex_str.upper()

    @staticmethod
    def file_path(file_path: FilePath) -> Path:
        """Validate a file path and return its normalized form.

        Args:
            file_path (FilePath): The file path to validate.

        Returns:
            Path: The normalized file path if valid.

        Raises:
            TypeError: If the file path is not a string or a Path object.
        """

        if isinstance(file_path, str):
            return Path(file_path)
        if isinstance(file_path, Path):
            return file_path
        raise TypeError("The file path must be a string or a Path object.")



# -*- encoding: utf-8 -*-
"""Unifont Utils - Converter"""

from functools import reduce
from typing import List


class Converter:
    """Class for converters."""

    @staticmethod
    def to_hex(data: List[int]) -> str:
        """Convert glyph pixel data to Unifont `.hex` format string.

        Args:
            data (List[int]): Glyph pixel data in the image, stored as `0` and `1`.

        Returns:
            str: The Unifont `.hex` format string.

        Raises:
            ValueError: If the glyph data is empty.
        """

        if not data:
            raise ValueError(
                "Unable to convert to .hex string. The glyph data is empty."
            )

        n = reduce(lambda acc, pixel: (acc << 1) | (1 if pixel else 0), data, 0)
        return hex(n)[2:].upper().zfill(32 if len(data) == 128 else 64)

    @staticmethod
    def to_img_data(hex_str: str, width: int = 16, height: int = 16) -> List[int]:
        """Convert Unifont `.hex` format string to glyph pixel data.

        Args:
            hex_str (str): The Unifont `.hex` format string.
            width (int, optional): The width of the glyph in pixels. Defaults to `16`.
            height (int, optional): The height of the glyph in pixels. Defaults to `16`.

        Returns:
            List[int]: The glyph pixel data in the image, stored as `0` and `1`.
        """

        if not hex_str:
            return []

        return [(int(hex_str, 16) >> i) & 1 for i in range(width * height - 1, -1, -1)]
